import numpy as np
from scipy.spatial import Delaunay

def find_edges_with(i, edge_set):
    return (
             [j for (x,j) in edge_set if x==i],
             [j for (j,x) in edge_set if x==i]
    )

def stitch_boundaries(edges):
    edge_set = edges.copy()
    boundary_lst = []
    while len(edge_set) > 0:
        boundary = []
        edge0 = edge_set.pop()
        boundary.append(edge0)
        last_edge = edge0
        while len(edge_set) > 0:
            i,j = last_edge
            j_first, j_second = find_edges_with(j, edge_set)
            if j_first:
                edge_set.remove((j, j_first[0]))
                edge_with_j = (j, j_first[0])
                boundary.append(edge_with_j)
                last_edge = edge_with_j
            elif j_second:
                edge_set.remove((j_second[0], j))
                edge_with_j = (j, j_second[0])  # flip edge rep
                boundary.append(edge_with_j)
                last_edge = edge_with_j

            if edge0[0] == last_edge[1]:
                break

        boundary_lst.append(boundary)
    return boundary_lst

def remove_collinear_points(points, tol=2e-4):
    """
    Remove points that are collinear with their neighbors in a polygon.

    Parameters:
        points: list of (x, y) tuples defining the polygon vertices.
                The polygon is assumed to be closed (first and last points are connected).
        tol: tolerance to decide if the area of the triangle is effectively zero.
    
    Returns:
        A new list of points with collinear points removed.
    """
    if len(points) < 3:
        return points.copy()

    # Check if the polygon is closed (i.e., first equals last)
    closed = np.array_equal(points[0], points[-1])
    # Work with the non-repeated version if the polygon is closed.
    pts = points[:-1] if closed else points.copy() #np.unique(, axis=0)

    new_points = []
    n = len(pts)
    total_area = 0.5 * abs(sum(x0 * y1 - x1 * y0 for (x0, y0), (x1, y1) in zip(points, points[1:] + [points[0]])))

    # Loop over each point, treating the polygon as cyclic.
    for i in range(n):
        prev = pts[(i - 1) % n]
        curr = pts[i]
        nxt = pts[(i + 1) % n]
        # Compute twice the area of the triangle
        area = abs(prev[0] * (curr[1] - nxt[1]) +
                   curr[0] * (nxt[1]  - prev[1]) +
                   nxt[0]  * (prev[1] - curr[1]))*0.5
        # Only keep the current point if the area is larger than tol.
        if area/total_area > tol:
            new_points.append(curr)

    # Re-close the polygon if necessary.
    if closed and new_points:
        new_points.append(new_points[0])
    return np.array(new_points)

def _add_edge(edges, i, j, only_outer=True):
    """
    Add an edge between the i-th and j-th points,
    if not in the list already
    """
    if (i, j) in edges or (j, i) in edges:
        # already added
        assert (j, i) in edges, "Can't go twice over same directed edge?"
        if only_outer:
            # if both neighboring triangles are in shape, it's not a boundary edge
            edges.remove((j, i))
        return
    edges.add((i, j))


def _alpha_default(points, tri) -> float:
    """
    Compute a simple 'default' alpha for an alpha-shape,
    assuming the Delaunay triangles are roughly equilateral.

    :param points: (N,2) array of 2D point locations
    :param tri: A scipy.spatial.Delaunay object built from `points`
    :return: A float alpha
    """

    # tri.simplices is an (M,3) array of indices into `points`
    simplices = tri.simplices

    # Collect all edges by pairs of indices from each simplex
    # Note: each triangle contributes edges (i,j), (j,k), (k,i)
    edges = []
    for simplex in simplices:
        i, j, k = simplex
        edges.append((i, j))
        edges.append((j, k))
        edges.append((k, i))

    # Convert edges list to a unique set so we don't repeat edges
    # (small performance optimization)
    edges = set(tuple(sorted(edge)) for edge in edges)

    # Compute lengths of all edges
    lengths = []
    for (i, j) in edges:
        pi = points[i]
        pj = points[j]
        lengths.append(np.linalg.norm(pi - pj))

    lengths = np.array(lengths)
    # We could use the mean, but here we use median for robustness
    median_length = np.median(lengths)

    # Equilateral triangle of side length a has circumscribed circle radius a/sqrt(3).
    # We'll just pick alpha ~ median_length / sqrt(3).
    alpha = median_length / np.sqrt(3)

    return 1/alpha

def _max_pairwise_distance(points: np.ndarray) -> float:
    """
    Return the maximum Euclidean distance between any two points
    in a set of 2D points.

    :param points: (n, 2) array of 2D points
    :return: The largest Euclidean distance between any pair
    """

    from scipy.spatial.distance import cdist
    # Edge case: if there is 0 or 1 point, max distance is 0
    if points.shape[0] < 2:
        return 0.0

    # Compute all pairwise distances
    dist_matrix = cdist(points, points, 'euclidean')

    # Return the maximum distance
    return float(dist_matrix.max())

def alpha_shape(points, alpha=None, only_outer=True, radius=None, bound_ratio=None, tri=None):
    """
    Compute the alpha shape (concave hull) of a set of points.
    :param points: np.array of shape (n,2) points.
    :param alpha: alpha value.
    :param only_outer: boolean value to specify if we keep only the outer border
    or also inner edges.
    :return: set of (i,j) pairs representing edges of the alpha-shape. (i,j) are
    the indices in the points array.
    """
    assert points.shape[0] > 3, "Need at least four points"

    if tri is None:
        tri = Delaunay(points)

    if bound_ratio is not None:
        assert radius is None
        assert alpha is None
        radius = _max_pairwise_distance(points) * bound_ratio
    if radius is not None:
        alpha = 1/radius
    if alpha is None:
        alpha = _alpha_default(points, tri)

    edges = set()
    # Loop over triangles:
    # ia, ib, ic = indices of corner points of the triangle
    for ia, ib, ic in tri.simplices:
        pa = points[ia]
        pb = points[ib]
        pc = points[ic]
        # Computing radius of triangle circumcircle
        # www.mathalino.com/reviewer/derivation-of-formulas/derivation-of-formula-for-radius-of-circumcircle
        a = np.sqrt((pa[0] - pb[0]) ** 2 + (pa[1] - pb[1]) ** 2)
        b = np.sqrt((pb[0] - pc[0]) ** 2 + (pb[1] - pc[1]) ** 2)
        c = np.sqrt((pc[0] - pa[0]) ** 2 + (pc[1] - pa[1]) ** 2)
        s = (a + b + c) / 2.0
        area = np.sqrt(s * (s - a) * (s - b) * (s - c))
        circum_r = a * b * c / (4.0 * area)
        if circum_r < 1/alpha:
            _add_edge(edges, ia, ib, only_outer=only_outer)
            _add_edge(edges, ib, ic, only_outer=only_outer)
            _add_edge(edges, ic, ia, only_outer=only_outer)

    # return points[stitch_boundaries(edges)]
    # return stitch_boundaries(edges)
    # return edges
    first_bound = stitch_boundaries(edges)[0]
    # return np.array(list(reversed([points[i] for edge in first_bound for i in edge])))
    ordered_points = np.array([points[i] for edge in first_bound for i in edge])
    _, idx = np.unique(ordered_points, return_index=True, axis=0)

    ordered_unique_points = ordered_points[np.sort(idx)]
    return remove_collinear_points(ordered_unique_points)
    # return ordered_points





import numpy as np
Array = np.ndarray
from veux.frame._section import SectionGeometry

import shps.rotor as so3

def _elastic_curve(x: Array, v: list, L:float, tangent=False)->Array:
    "compute points along Euler's elastica"
    ui, vi, uj, vj = v

    xi = x/L                        # local coordinate
    if not tangent:
        N1 = 1.-3.*xi**2+2.*xi**3
        N2 = L*(xi-2.*xi**2+xi**3)
        N3 = 3.*xi**2 - 2*xi**3
        N4 = L*(xi**3 - xi**2)
        y = ui*N1 + vi*N2 + uj*N3 + vj*N4
        return y.flatten()

    else:
        M3 = 1 - xi
        M4 = 6/L*(xi-xi**2)
        M5 = 1 - 4*xi+3*xi**2
        M6 = -2*xi + 3*xi**2
        return (ui*M3 + vi*M5 + uj*M4 + vj*M6).flatten()


def _hermite_cubic(
        xi,
        coord: Array,
        u: Array = None,     #: Displacements at two nodes
        v: Array = None,     #: Rotation vectors at two nodes
        Q = None,
        tangent: bool = False
    ):

    # Element length
    L = np.linalg.norm(coord[-1] - coord[0])

    (li, ti, vi), (lj, tj, vj) = u
    (si, ei, pi), (sj, ej, pj) = v

    Lnew  = L + lj - li
    xaxis = np.array([xi*Lnew])

    plan_curve = _elastic_curve(xaxis, [ti, pi, tj, pj], Lnew)
    elev_curve = _elastic_curve(xaxis, [vi,-ei, vj,-ej], Lnew)

    local_curve = np.stack([xaxis + li, plan_curve, elev_curve])

    return Q.T@local_curve + coord[0][None,:].T



class _FrameState:
    pass

class _GaussSample:
    def __init__(self, index):
        self.index = index

class _SimpleSample:
    def __init__(self, index):
        self.index = index

class _FrameElement:
    def __init__(self, 
                 tag, 
                 vmodel, 
                 xmodel=None, 
                 samples: int = None,
                 section_override=None,
                 interpolation=None):

        if samples is None:
            samples = len(vmodel.cell_nodes(tag))

        self._tag = tag
        self._vmodel = vmodel
        self._xmodel = xmodel
        self._section_override = section_override
        self._elem_data = vmodel.cell_properties(tag)

        # Set section source
        section_source = "Default"
        if self._tag in self._vmodel._frame_outlines:
            section_source = "External"
        elif "sections" in self._elem_data:
            section_source = "Internal"
        
        self._section_source = section_source

        #

        # set _gauss_sample_points
        self._gauss_sample_points = []
        if self._xmodel is not None:
            self._gauss_sample_points = self._xmodel.eleResponse(self._tag, "integrationPoints")

        self._gauss_sections = [
            self._vmodel._frame_section(tag)
            for tag in self._elem_data.get("sections", [])
        ]    

        # Simple samples
        if samples is not None:
            self._simple_sample_points = np.linspace(0, 1.0, samples)
        elif section_source == "External":
            self._simple_sample_points = np.linspace(0, 1.0, len(self._vmodel._frame_outlines[self._tag]))
        elif section_source == "Internal":
            self._simple_sample_points = np.linspace(0, 1.0, len(self._elem_data["sections"]))
        else:
            self._simple_sample_points = np.linspace(0, 1.0, 2)

        # Interpolate section geometry
    


    def sample_section(self, sample)-> "SectionGeometry":
        if self._section_override is not None:
            return self._section_override

        if self._section_source == "External":
            return self._vmodel._frame_outlines[self._tag][sample.index]

        elif self._section_source == "Internal":
            return self._vmodel._frame_section(self._elem_data["sections"][0])#sample.index])

        else: 
            return self._vmodel._frame_section(None)


    def simple_samples(self):
        return (_SimpleSample(i) for i in range(len(self._simple_sample_points)))

    # def gauss_samples(self):
    #     return (_GaussSample(i)  for i in range(len(self._gauss_sample_points)))

    # Create states
    def gauss_state(self, state=None, positions=None, rotations=None):
        """
        Create a FrameState
        """

    def nodal_state(self, state=None, positions=None, rotations=None):
        """
        Create a FrameState
        """
        ...


    def sample_position(self, sample, state=None):
        xs = self._simple_sample_points[sample.index]
        model = self._vmodel
        Xi, Xj = self._vmodel.cell_position(self._tag)[[0,-1],:]
        X = Xi,Xj
        Xs = Xi + (
            (Xj - Xi) * self._simple_sample_points[sample.index]
        )

        if state is None:
            return Xs

        Q = model.frame_orientation(self._tag)
        u = [
            xi - X[i] for i,xi in enumerate(model.cell_position(self._tag, state=state)[[0,-1],:])
        ]
        R = [
            model.node_rotation(node, state=state) for node in model.cell_nodes(self._tag)
        ]
        x = _hermite_cubic(xs, X,
                Q=Q,
                u = [Q@u[0],          Q@u[1]],
                v = [Q@so3.log(R[0]), Q@so3.log(R[-1])]).T
        return x[0]

    def sample_rotation(self, sample, state=None):
        xs = self._simple_sample_points[sample.index]
        Rs = self._vmodel.frame_orientation(self._tag).T

        if state is None:
            return Rs
        
        model = self._vmodel
        Ri,*_,Rj = [
            model.node_rotation(node, state=state) for node in model.cell_nodes(self._tag)
        ]
        return so3.exp(xs*so3.log(Rj@Ri.T))@Ri@Rs

    def sample_material(self, sample, state=None):
        pass

    def sample_weight(self, sample):
        pass




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Copyright (c) 2025, Claudio M. Perez
# All rights reserved.  No warranty, explicit or implicit, is provided.
#
# This source code is licensed under the BSD 2-Clause License.
# See LICENSE file or https://opensource.org/licenses/BSD-2-Clause
#
from .frame import Motion, skin_frames


import numpy as np
from scipy.spatial.transform import Rotation

import pygltflib

from veux.canvas.gltf import GLTF_T


def _append_index(lst, item):
    lst.append(item)
    return len(lst) - 1

def _create_mesh(canvas,
                  positions,
                  texcoords,
                  joints_0,
                  weights_0,
                  indices,
                  skin_idx=None,
                  material=None):
    
    gltf = canvas.gltf
    joints_0  = np.array(joints_0,  dtype=canvas.index_t)
    weights_0 = np.array(weights_0, dtype=canvas.float_t)
    indices   = np.array(indices,   dtype=canvas.index_t).reshape(-1)

    jnt_bytes = joints_0.tobytes()
    wts_bytes = weights_0.tobytes()

    # Accessors
    positions = np.array(positions, dtype=canvas.float_t)
    ver_accessor = _append_index(gltf.accessors, pygltflib.Accessor(
        bufferView=canvas._push_data(positions.tobytes(), pygltflib.ARRAY_BUFFER),
        componentType=GLTF_T[canvas.float_t],
        count=len(positions),
        type="VEC3",
        min=positions.min(axis=0).tolist(),
        max=positions.max(axis=0).tolist()
    ))

    texcoords = np.array(texcoords, dtype=canvas.float_t)
    tex_accessor = _append_index(gltf.accessors, pygltflib.Accessor(
        bufferView=canvas._push_data(texcoords.tobytes(), pygltflib.ARRAY_BUFFER),
        componentType=GLTF_T[canvas.float_t],
        count=len(texcoords),
        type="VEC2"
    ))

    jnt_accessor = _append_index(gltf.accessors, pygltflib.Accessor(
        bufferView=canvas._push_data(jnt_bytes, pygltflib.ARRAY_BUFFER),
        componentType=GLTF_T[canvas.index_t],
        count=len(joints_0),
        type="VEC4"
    ))

    wts_accessor = _append_index(gltf.accessors, pygltflib.Accessor(
        bufferView=canvas._push_data(wts_bytes, pygltflib.ARRAY_BUFFER),
        componentType=GLTF_T[canvas.float_t],
        count=len(weights_0),
        type="VEC4"
    ))


    idx_accessor = _append_index(gltf.accessors, pygltflib.Accessor(
        bufferView=canvas._push_data(indices.tobytes(), pygltflib.ELEMENT_ARRAY_BUFFER),
        componentType=GLTF_T[canvas.index_t],
        count=len(indices),
        type="SCALAR",
        min=[int(indices.min())],
        max=[int(indices.max())]
    ))

    # Create the Mesh
    mesh = pygltflib.Mesh(
        primitives=[
            pygltflib.Primitive(
                attributes=pygltflib.Attributes(
                    POSITION=ver_accessor,
                    JOINTS_0=jnt_accessor,
                    WEIGHTS_0=wts_accessor,
                    TEXCOORD_0=tex_accessor
                ),
                material=material,
                indices=idx_accessor,
                mode=pygltflib.TRIANGLES
            )
        ],
        name="FrameSkinMesh"
    )

    mesh_idx = _append_index(gltf.meshes, mesh)

    #
    # 4) Create a Node referencing the mesh + skin
    #
    mesh_node_idx = _append_index(gltf.nodes, pygltflib.Node(
        mesh=mesh_idx,
        skin=skin_idx,
        name="FrameSkinMeshNode"
    ))

    # Put it in the scene
    if not gltf.scenes or len(gltf.scenes)==0:
        gltf.scenes = [pygltflib.Scene(nodes=[])]

    gltf.scenes[0].nodes.append(mesh_node_idx)



def deform_extrusion(model, canvas, state, skin_nodes, config=None):
    """
    Given a 'state' that contains the updated (displaced/rotated) coordinates 
    for each element’s cross section,
    update the glTF nodes' translation/rotation accordingly.

    The skinned mesh in glTF will automatically show the new shape
    as the viewer or engine processes the node transforms.
    """
    gltf = canvas.gltf 

    if config is None:
        config = {}

    for element_name, el in model["assembly"].items():

        # Number of cross sections
        nen = len(el["nodes"])
        # TODO: Make these consistent with draw_sections
        # Displacements and rotations from 'state' for each cross-section
        pos_all = state.cell_array(element_name, state.position)  # shape (nen, 3?)
        rot_all = state.cell_array(element_name, state.rotation)  # shape (nen, 3x3) ?

        # Original coordinates
        X_ref = np.array(el["crd"])  # shape (nen, 3)

        for j in range(nen):
            # Look up the node index in glTF
            if (element_name, j) not in skin_nodes:
                continue
            node_idx = skin_nodes[(element_name, j)]


            gltf.nodes[node_idx].translation = (X_ref[j,:] + pos_all[j,:]).tolist()
            gltf.nodes[node_idx].rotation = [*Rotation.from_matrix(rot_all[j] ).as_quat()]


def create_animation(artist, states=None, skin_nodes=None):

    # 2) Create the animation helper
    animation = Motion(artist)

    # 3) For each state, record a new keyframe
    for time in states.times:
        animation.draw_sections(state=states[time])
        animation.advance()

    animation.add_to(artist.canvas)
    return animation


def _animate(sam_file, res_file=None, vertical=None, **opts):
    # Configuration is determined by successively layering
    # from sources with the following priorities:
    #      defaults < file configs < kwds
    from veux.frame import FrameArtist
    import veux.canvas.gltf
    from veux.model import read_model

    config = veux.config.Config()


    if sam_file is None:
        raise RenderError("Expected positional argument <sam-file>")

    if isinstance(sam_file, (str,)):
        model = read_model(sam_file)

    elif hasattr(sam_file, "asdict"):
        # Assuming an opensees.openseespy.Model
        model = sam_file.asdict()

    elif hasattr(sam_file, "read"):
        model = read_model(sam_file)


    if "RendererConfiguration" in model:
        veux.apply_config(model["RendererConfiguration"], config)

    veux.apply_config(opts, config)
    if vertical is not None:
        config["artist_config"]["vertical"] = vertical

    artist = FrameArtist(model, ndf=6,
                         config=config["artist_config"],
                         model_config=config["model_config"],
                         canvas=veux.canvas.gltf.GltfLibCanvas())


    if res_file is not None:
        if isinstance(res_file, str):
#           soln = artist.model.wrap_state(res_file)
            soln = veux.model.read_state(res_file, artist.model, **opts["state_config"])
        else:
            from veux.state import GroupSeriesSE3, StateSeries
            series = StateSeries(res_file, artist.model,
                transform = artist.dofs2plot,
                recover_rotations="conv"
            )
            soln = GroupSeriesSE3(series, artist.model, recover_rotations="conv", transform=artist.dofs2plot)

        if "time" not in opts:
            create_animation(artist, soln)
        else:
            skin_nodes,_ = skin_frames(artist.model, 
                                       artist.canvas,
                                        config=artist._config_sketch("default")["surface"]["frame"])
            deform_extrusion(artist.model, artist.canvas, soln, skin_nodes)

    return artist


if __name__ == "__main__":
    import sys
    from veux.errors import RenderError
    import veux.parser
    config = veux.parser.parse_args(sys.argv)

    try:
        artist = _animate(**config)

        # write plot to file if output file name provided
        if config["write_file"]:
            artist.save(config["write_file"])

        # Otherwise either create popup, or start server
        elif hasattr(artist.canvas, "popup"):
            artist.canvas.popup()

        elif hasattr(artist.canvas, "to_glb"):
            from veux.server import Server
            from veux.viewer import Viewer
            viewer = Viewer(artist, viewer=config["viewer_config"].get("name", None))
            port = config["server_config"].get("port", None)
            server = Server(viewer=viewer)
            server.run(port=port)

        elif hasattr(artist.canvas, "to_html"):
            import veux.server
            port = config["server_config"].get("port", None)
            server = veux.server.Server(html=artist.canvas.to_html())
            server.run(port=port)

    except (FileNotFoundError, RenderError) as e:
        # Catch expected errors to avoid printing an ugly/unnecessary stack trace.
        print(e, file=sys.stderr)
        print("         Run '{NAME} --help' for more information".format(NAME=sys.argv[0]), file=sys.stderr)
        sys.exit()




#===----------------------------------------------------------------------===#
#
#         STAIRLab -- STructural Artificial Intelligence Laboratory
#
#===----------------------------------------------------------------------===#
#
# Claudio Perez
# Summer 2024
#
import sys
import socket
import threading
from wsgiref.simple_server import make_server
import bottle



def _open_borderless_viewer(url):
    # TODO
    from playwright.sync_api import sync_playwright
    with sync_playwright() as p:
        browser = p.chromium.launch(
            headless=False,
            args=[
                '--app=' + url,     # launch in app mode
                '--window-size=800,600',
                '--disable-infobars',
                '--no-default-browser-check',
                '--disable-extensions',
            ]
        )
        context = browser.new_context()
        context.new_page()  # required to keep browser alive on some systems
        input("Press Enter to exit and close browser...")
        browser.close()

def _check_port(port):
    # Check if the port is available
    with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
        try:
            s.bind(("localhost", port))
            return True
        except OSError as e:
            return False

def _find_port():
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(("",0))
    s.listen(1)
    port = s.getsockname()[1]
    s.close()
    return port

class Server:
    def __init__(self, viewer=None, html=None):
        self._app = bottle.Bottle()
        self._server = None

        if html is not None:
            self._app.route("/")(lambda : html )
            return

        if viewer is not None:

            html = viewer.get_html()
            self._app.route("/")(lambda : html )
            @self._app.route("/quit")
            def _quit():
                threading.Thread(target=self._shutdown).start()
                return "Shutting down..."


            for path, page in viewer.resources():
                self._app.route(path)(page)

    def _shutdown(self):
        if self._server:
            self._server.shutdown()

    def run(self, port=None):
        if port is None:
            # Default to something consistent
            port = 8081 if _check_port(8081) else _find_port()

        print(f"  Displaying at http://localhost:{port}/ \n  Press Ctrl-C to quit.\n")

        self._server = make_server("localhost", port, self._app)
        self._server.serve_forever()


if __name__ == "__main__":

    options = {
        "viewer": None
    }
    argi = iter(sys.argv[1:])

    for arg in argi:
        if arg == "--viewer":
            options["viewer"] = next(argi)
        else:
            filename = arg

    with open(filename, "rb") as f:
        glb = f.read()

    Server(glb=glb, **options).run()
